Appendix A

# R script for running random forest algorithm on multi layer raster file
# and producing cross-validated accuracy assessment
# Author: Koreen Millard, Department of Geography and Environmental Studies
# Carleton University, Ottawa ON, Canada
# Correspondence: Koreen Millard (koreen_millard@carleton.ca)
# last updated October 13, 2015
# Please acknowledge the use of this script in any publication
# Millard and Richardson, Remote Sens. 2015, 7(7), 8489-8515; doi:10.3390/rs70708489
# an example dataset has been provided for testing 


# If not installed, must install all relevant libraries first
#install.packages("raster", "randomForest", "sp", "rgdal")

#load libraries
library(raster)
library(randomForest)
library(sp)
library(rgdal)

setwd("C:\\Data\\") #set up working directory - this is where all data is stored

infile = "Composite.tif" 
#input raster will take any multi-channel image types (pix, tif, etc)

AllPts<-read.csv("C:\\Data\\TrainingData.csv", header=TRUE, sep = ",")  
# csv containing independent training data points.   
# (XY location for each individual point)
# this csv must have a field/column for Easting (POINT_X) and Northing (POINT_Y) 
# and a field "ClassID" representing unique classes (can be either text or integer)
# or change these fields in the script below

NumToExcludeForValidation<-20  
#alter this as you require
#defines the number of validation points to be extracted from AllPts
NumberForTraining<-48    
#defines the number of training points to be extracted from AllPts
NumIterations<-25 
#number of times to run the classification
selection<-c(1:6) 
#selection of channels in image you want to use

#first subset training data to remove the points for validation points
index<-1:nrow(AllPts)
ValidationIndex<-sample(index, NumToExcludeForValidation)
NotValidationPts<-AllPts[-ValidationIndex,]
ValidationPts<-AllPts[ValidationIndex,]
coordinates(ValidationPts)=~POINT_X+POINT_Y 
proj4string(ValidationPts)<-CRS("+proj=utm +zone=18 +datum=WGS84 +units=m +no_defs +ellps=WGS84 +towgs84=0,0,0") 
#must define coordinate reference system appropriately

cumulative_errorTable = NULL    
#this variable will hold the error table populated in subsequent iterations of the loop

#loops are slow in R, but easy to implement
for (i in 1:NumIterations){
  cat("Running classification iteration ", i)
  #now get the subset of training data
  index2<-1:nrow(AllPts)
  TrainingIndex<-sample(index2, NumberForTraining)
  TrainingPts<-na.omit(NotValidationPts[TrainingIndex,])
  coordinates(TrainingPts)=~POINT_X+POINT_Y
  proj4string(TrainingPts)<-CRS("+proj=utm +zone=18 +datum=WGS84 +units=m +no_defs +ellps=WGS84 +towgs84=0,0,0")  
 #must define coordinate reference system appropriately
  
  # Extract the training data 
  inraster = stack(infile)
  training_data = extract(inraster, TrainingPts)
  training_response = as.factor(TrainingPts$ClassID) 
  #here, our csv file has a field called "ClassID" for names/code of each of the classes 

  training_predictors = training_data[,selection]  
  #here we subset to only use the selected channels (defined above)

  # Create and save the forest
  ntree = 1000     #number of trees to produce per iteration
  r_forest = randomForest(training_predictors, y=training_response, ntree = ntree, keep.forest=TRUE, importance = TRUE) 
  #displays graph and list for all classes (several options)
  imp<-importance(r_forest)  #for ALL classes individually
  
  
  r_forest	#display out of bag error and confusion matrix in console
  imp   #display importance output in console
  
  #set up the file name for the raster to print
  outraster = paste("rf_iteration", i, "NumberForTraining", NumberForTraining, ".pix", sep = "_")
 
  # Classify the whole image
  predictor_data = subset(inraster, selection)	
  #gets the raster data to use for classification
  setwd("C:\\data\\") #change this to your output directory if different 

  predictions = predict(predictor_data, r_forest, filename=outraster, format="PCIDSK", overwrite=TRUE, progress="text", type="response") #write out the predicted classification to a raster file
  remove(predictor_data) 
  
  #do independent validation
  outraster=stack(outraster)	 
  #reads in the classification you just produced
  validation=extract(outraster, ValidationPts)  
  #extracts the value of the classified raster at the validation point locations

  ValidationJoin<-cbind(ValidationPts$ClassID, validation) 
  #joins the predicted with observed

  totalcorrect<-(subset(ValidationJoin,ValidationJoin[,1]==ValidationJoin[,2])) 
  #determines which samples were the same for observed and predicted

  indep_errorrate<-100 - (length(totalcorrect[,1])/length(validation)*100)
  #determines the independent error rate (%)

  rfOOB<-r_forest$err.rate[ntree,1]*100 #obtains the rfOOB error rate
  differror<-rfOOB - indep_errorrate
  current_error<-cbind(names(outraster), rfOOB, indep_errorrate, differror)
  #binds all rfOOB, indep_errorrate, and differror together

  cumulative_errorTable<-rbind(current_error, cumulative_errorTable)	
  # populates a variable of cumulatively collected validation data within the loop

} #end of loop

tableName = paste("rf_iteration", i,"NumberForTraining", NumberForTraining, ".csv", sep = "_")
write.csv(cumulative_errorTable, file = tableName)	
#prints out a table with error from all classifications

